import logging
import sys

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[
        logging.StreamHandler(sys.stdout)
    ]
)

if not hasattr(sys, 'frozen'):
    import os
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))

import otter

otter.main()



# TODO: consider storing database in module data:
# import importlib.resources as resources
# with resources.path("otter.data", "otter.db") as db_path:

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

from otter.const import OTTER_DB


def engine():
    return create_engine(f"sqlite:///{OTTER_DB}")


def session() -> Session:
    s = sessionmaker()
    s.configure(bind=engine())
    return s()



from sqlalchemy.exc import NoResultFound

import otter.database.query as db
from otter import database
from otter.database.model import Project, ProjectTask, Record
from otter.model import OtterRecordSyncCalculator
from otter.odoo import rest
from otter.util.date import work_date


def get_or_create_record(project_task_id):
    """
    :param project_task_id: project task ID
    :return: record
    """

    current_date = work_date()
    session = database.session()

    try:
        return (
            session.query(
                Record
            )
            .filter(Record.date == current_date)
            .filter(Record.project_task_id == project_task_id)
            .filter(Record.odoo_id.is_(None))  # only unsynced
            .one()
        )
    except NoResultFound:
        record = db.add_record(project_task_id, session=session, add_date=current_date)
        session.commit()
        session.refresh(record)
        return record


def get_records(query_date, session=database.session()):
    return (
        session.query(
            Record
        )
        .filter(Record.date == query_date)
        .all()
    )


def sync_remote_projects(session=database.session()):
    projects_json = rest.get_projects_json()
    if projects_json is not None:
        for project_json in projects_json:
            session.merge(Project.from_odoo_json(project_json))
    session.commit()


def sync_remote_project_tasks(session=database.session()):
    project_tasks_json = rest.get_project_tasks_json()
    if project_tasks_json is not None:
        for task_json in project_tasks_json:
            session.merge(ProjectTask.from_odoo_json(task_json))
    session.commit()


def sync_local_records(query_date, session=database.session()):
    records = (
        session.query(
            Record
        )
        .filter(Record.date <= query_date)
        .filter(Record.odoo_id.is_(None))  # only unsynced
        .all()
    )

    if len(records) > 0:
        record_sync_calc = OtterRecordSyncCalculator(records)

        for i, day in enumerate(record_sync_calc.dates()):
            # print() if i > 0 else None
            # print(f"{date_to_weekday_string(day)}, {date_to_date_string(day)}:")
            sync_records = record_sync_calc.records(day)

            for sync_record in sync_records:
                assert sync_record.odoo_id is None

                if sync_record.duration == 0:
                    session.delete(sync_record)
                    session.commit()
                    continue

                record_odoo_id = rest.post_record(sync_record.odoo_create_json())
                assert record_odoo_id is not None
                sync_record.odoo_id = record_odoo_id
                session.commit()


def sync_remote_records(session=database.session()):
    record_json = rest.get_records_json()
    if record_json is not None:
        for record_json in record_json:
            local_id = Record.get_id_by_odoo_id(record_json['id'], session)
            session.merge(Record.from_odoo_json(local_id, record_json))
    session.commit()



from otter.util.math import ceil_duration, float_mod, round_duration, floor_duration


class OtterRecordSyncCalculator(object):

    def __init__(self, records=None):
        self.records_by_date = {}
        if records is None:
            records = []
        self.add_records(records)

    def add_record(self, record):
        if record.date not in self.records_by_date:
            self.records_by_date[record.date] = []
        self.records_by_date[record.date].append(record)

    def add_records(self, records):
        for record in records:
            self.add_record(record)

    def dates(self):
        return self.records_by_date.keys()

    def target_duration(self, date):
        """
        Calculate the overall duration of all records of a day (date)
        and round up the result to the next 15 minutes (0.25 in decimal).
        E.g. a duration of 7.83 will be rounded up to 8.0.
        """

        duration_hours = 0.0
        for record in self.records_by_date[date]:
            duration_hours += record.duration
        return ceil_duration(duration_hours)

    def records(self, date):
        """
        Get the records of a day (date) with re-calculated durations.
        """

        records = []

        # this is the target value to reach
        target_duration_hours = self.target_duration(date)

        # the sum of all rounded records can be lower or higher than the target value,
        # this is what we are going to fix here
        duration_hours_records_rounded = 0.0
        for record in self.records_by_date[date]:
            duration_hours_records_rounded += round_duration(record.duration)

        # a negative difference means that the summed up rounded durations
        # result in too high overall duration, we need to reduce it
        difference = target_duration_hours - duration_hours_records_rounded

        # how many times we have to fix the rounded value?
        times = int(abs(difference // 0.25))
        mod_record_tuples = []

        for record in self.records_by_date[date]:
            mod = float_mod(record.duration * 4.0) / 4.0
            mod_record_tuples.append((mod, record))

        increase = difference > 0
        mod_record_tuples.sort(key=lambda tup: tup[0], reverse=increase)

        for mod, record in mod_record_tuples:
            target_duration_hours = round_duration(record.duration)
            ceiled = record.duration < target_duration_hours

            if times > 0:
                if increase and not ceiled:
                    target_duration_hours = ceil_duration(record.duration)
                    times = times - 1
                elif not increase and ceiled:
                    target_duration_hours = floor_duration(record.duration)
                    times = times - 1

            # set the fixed values for name and unit_amount
            record.name = record.description
            record.unit_amount = target_duration_hours
            records.append(record)

        return records



from __future__ import print_function
import sys
from pathlib import Path

from otter.const import OTTER_DIR, OTTER_CFG

# Paths must be initiated here to be available for all modules
Path(OTTER_DIR).mkdir(parents=True, exist_ok=True)
Path(OTTER_CFG).touch(exist_ok=True)


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


# from otter.otter import main

# __all__ = ['main']



import requests
import sys

from otter.config import deserialize
from otter.odoo import cookies


def login(url, db, username, password):
    request_json = {
        "jsonrpc": "2.0",
        "params": {
            "db": db,
            "login": username,
            "password": password
        }
    }

    response = requests.post(url + '/web/session/authenticate', json=request_json)

    # before: and 'result' in response.json():
    if response.status_code != 200 or 'session_id' not in response.cookies:
        print("ERROR DURING LOGIN")
        sys.exit(1)

    uid = response.json()['result']['uid']
    session = response.cookies['session_id']

    return uid, session


def get_databases_json(url):
    request_json = {
        "jsonrpc": "2.0",
        "context": {}
    }

    # endpoint only for Odoo version >= 10
    response = requests.get(url + '/web/database/list', json=request_json)

    if response.status_code != 200 or 'result' not in response.json():
        print("ERROR FETCHING DATABASES")
        sys.exit(1)

    return response.json()['result']


def get_projects_json():
    cfg = deserialize()

    url = cfg['url']

    request_json = {
        "jsonrpc": "2.0",
        "method": "call",
        "params": {
            "model": "project.project",
            "fields": [
                "id",
                "name",
                "active"
            ],
            "sort": "id ASC"
        }
    }

    response = requests.post(f"{url}/web/dataset/search_read", json=request_json, cookies=cookies())

    if response.status_code != 200 or 'result' not in response.json():
        print("ERROR")
        return None

    return response.json()['result']['records']


def get_project_tasks_json():
    cfg = deserialize()

    url = cfg['url']

    request_json = {
        "jsonrpc": "2.0",
        "method": "call",
        "params": {
            "model": "project.task",
            "domain": [
                ["project_id", "!=", False]
            ],
            "fields": [
                "id",
                "name",
                "active",
                "project_id"
            ],
            "sort": "id ASC"
        }
    }

    response = requests.post(f"{url}/web/dataset/search_read", json=request_json, cookies=cookies())

    if response.status_code != 200 or 'result' not in response.json():
        print("ERROR")
        return None

    return response.json()['result']['records']


def get_records_json():
    cfg = deserialize()

    url = cfg['url']
    uid = cfg['uid']

    request_json = {
        "jsonrpc": "2.0",
        "method": "call",
        "params": {
            "model": "account.analytic.line",
            "domain": [
                ["project_id", "!=", False],
                ["task_id", "!=", False],
                ["user_id", "=", uid]
            ],
            'date': '2021-01-04',
            "sort": "id ASC"
        }
    }

    response = requests.post(f"{url}/web/dataset/search_read", json=request_json, cookies=cookies())

    if response.status_code != 200 or 'result' not in response.json():
        print("ERROR")
        return None

    return response.json()['result']['records']


def post_record(record):
    cfg = deserialize()

    url = cfg['url']
    uid = cfg['uid']

    request_json = {
        "jsonrpc": "2.0",
        "method": "call",
        "params": {
            "args": [
                {
                    "date": str(record['date']),
                    "unit_amount": record['duration'],
                    "name": record['names'],
                    "project_id": record['project_id'],
                    "task_id": record['task_id']
                }
            ],
            "model": "account.analytic.line",
            "method": "create",
            "kwargs": {
                "context": {
                    "uid": uid,
                }
            }
        }
    }

    response = requests.post(f"{url}/web/dataset/call_kw/account.analytic.line/create", json=request_json,
                             cookies=cookies())

    if response.status_code != 200 or 'result' not in response.json():
        print("ERROR")
        return None

    return response.json()['result']



#!/usr/bin/env python

from setuptools import find_packages, setup

from otter.const import OTTER_PACKAGE, OTTER_DESCRIPTION, OTTER_VERSION

with open("README.md", "r") as file:
    long_description = file.read()

requirements = []
with open("requirements.txt", "r") as file:
    for line in file.read().splitlines():
        if not line.strip().startswith("#"):
            requirements.append(line)

setup(
    name=OTTER_PACKAGE,
    version=OTTER_VERSION,
    description=OTTER_DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/pathob/odoo-otter',
    author='Patrick Hobusch',
    author_email='pathob@users.noreply.github.com',
    license='Apache 2.0',
    packages=find_packages(),
    zip_safe=False,

    install_requires=requirements,

    entry_points={
        'console_scripts': [
            'otter=otter.otter:main'
        ],
    },
)
